
package book.jmx.examples;

import java.net.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import javax.management.*;
import javax.management.modelmbean.*;
import javax.management.loading.*;

   
public class XMBean implements ModelMBean, MBeanRegistration,
    XMBeanConstants {

  protected MBeanServer server = null;    
  
  // sequence numbers for notifications
  protected long notifierSequence      = 0;
  protected long attrNotifierSequence  = 0;
 
  // support for generic notification listeners
  private NotificationBroadcasterSupport notifier =
      new NotificationBroadcasterSupport();
        
  // maps mbean attribute and operation names to
  // corresponding XMBeanAttribute and XMBeanOperation   
  protected Map attributeMap          = new HashMap();
  protected Map operationMap          = new HashMap();    

  // managed resource, resource type, model mbean metadata
  // and object name reference
  protected Object resource           = null;
  protected String resourceType       = null;
  protected ModelMBeanInfo metadata   = null;
  protected ObjectName name           = null;
  
  
  // Constructors.
  
  public XMBean() {}
  
  public XMBean(ModelMBeanInfo info) throws MBeanException {  
    setModelMBeanInfo(info);
    
    try { 
      load();
    } catch (InstanceNotFoundException e) {}
  }
  
  public XMBean(String url, String resource) throws MBeanException {

    try {
      MetaDataBuilder builder = new XMLMetaDataBuilder(resource, url);
      setModelMBeanInfo(builder.build());
      setManagedResource(DefaultLoaderRepository.loadClass(resource)
          .newInstance(), OBJECT_REF);
        
      try {
        load();
      } catch (InstanceNotFoundException e) {}
    }
    catch (ClassNotFoundException e) {
      throw new MBeanException(
          e, "Unable to load class " + resource
      );
    }
    catch (InvalidTargetObjectTypeException e) {
      throw new Error("Invalid resource type 'ObjectReference'.");
    }
    catch (JMException e) {
      throw new MBeanException(e);
    }
    catch (Exception e) {
      e.printStackTrace();
      throw new MBeanException(e);
    }
  }
  
  
  // ModelMbean interface
    
  public void setModelMBeanInfo(ModelMBeanInfo metadata)
      throws MBeanException, RuntimeOperationsException {
  
    if (metadata == null) {
      throw new IllegalArgumentException(
          "The Model MBeanInfo cannot be null."
      );
    }
    
    if (metadata instanceof XMBeanInfo)
    try {
        metadata = new XMLMetaDataBuilder("", (String)metadata.getMBeanDescriptor().getFieldValue(XML_DEFINITION_URL)).build();
    } catch (Exception e) {
        e.printStackTrace();
        throw new MBeanException(e);
    }
      
    this.metadata = metadata;
  }
    
  public void setManagedResource(Object ref, String resourceType)
      throws MBeanException,
             InstanceNotFoundException,
             InvalidTargetObjectTypeException {

    if (ref == null) {
      throw new IllegalArgumentException(
          "Resource reference cannot be null."
      );
    }
      
    // check that is a supported resource type
    if (!isSupportedResourceType(resourceType)) {
      throw new InvalidTargetObjectTypeException(
          "Unsupported resource type: " + resourceType
      );
    }
    
    this.resource = ref;
    this.resourceType = resourceType;
  }

  protected boolean isSupportedResourceType(
      String resourceType) {
    
    if (resourceType == null)             return false;
    if (resourceType.equals(OBJECT_REF))  return true;

    return false;    
  }


  // ModelMBeanNotificationBroadcaster interface
  
  public void addNotificationListener(
      NotificationListener l, NotificationFilter filter,
      Object hback) {
      
    notifier.addNotificationListener(l, filter, hback);
  }
  
  public void removeNotificationListener(
      NotificationListener l) 
      throws ListenerNotFoundException {

    notifier.removeNotificationListener(l);
  }
  
  public void addAttributeChangeNotificationListener(
      NotificationListener l, String attributeName,
      Object hback) throws MBeanException {
          
    AttributeChangeNotificationFilter filter = 
        new AttributeChangeNotificationFilter();
        
    filter.enableAttribute(attributeName);
      
    notifier.addNotificationListener(l, filter,hback);
  }
  
  public void removeAttributeChangeNotificationListener(
      NotificationListener l, String attributeName)
      throws MBeanException, ListenerNotFoundException {
          
    notifier.removeNotificationListener(l);  
  }
  
  public void sendNotification(String message) 
      throws MBeanException {
      
    Notification notif = new Notification(
        GENERIC_MODELMBEAN_NOTIFICATION, // type
        this,                            // source
        ++notifierSequence,              // sequence number
        message                          // message
    );
    
    sendNotification(notif);
  }
  
  public void sendNotification(Notification notif)
      throws MBeanException {
      
    notifier.sendNotification(notif);
  }
  
  public void sendAttributeChangeNotification(
      AttributeChangeNotification notif)
      throws MBeanException {
          
    notifier.sendNotification(notif);
  }
  
  public void sendAttributeChangeNotification(
      Attribute oldValue, Attribute newValue)
      throws MBeanException {
          
    String attr = oldValue.getName();
    String type = oldValue.getClass().getName();
    
    AttributeChangeNotification notif = 
        new AttributeChangeNotification(
            this,                          // source
            ++attrNotifierSequence,        // seq. #
            System.currentTimeMillis(),    // time stamp
            "" + attr + " changed from "   // message
            + oldValue + " to " + newValue,
            attr, type,                    // name & type
            oldValue, newValue             // values
        );

    notifier.sendNotification(notif);        
  }
                      
  public MBeanNotificationInfo[] getNotificationInfo() {
    
    int size = metadata.getNotifications().length;
    MBeanNotificationInfo[] notifInfo = metadata.getNotifications();
    
    MBeanNotificationInfo[] modelInfo =
        new MBeanNotificationInfo[size + 2];
    
    for (int i = 0; i < size ;++i)
      modelInfo[i] = notifInfo[i];
    
    Descriptor descr1 = new DescriptorSupport();
    descr1.setField("name", "generic");
    descr1.setField("descriptorType", "notification");
    descr1.setField("severity", "5");
    
    ModelMBeanNotificationInfo generic = new ModelMBeanNotificationInfo(
        new String[] { GENERIC_MODELMBEAN_NOTIFICATION },
        "generic",
        "A generic Model MBean notification.",
        descr1
    );
    
    Descriptor descr2 = new DescriptorSupport();
    descr2.setField("name", AttributeChangeNotification.class.getName());
    descr2.setField("descriptorType", "notification");
    
    ModelMBeanNotificationInfo attrChange = new ModelMBeanNotificationInfo(
        new String[] { AttributeChangeNotification.ATTRIBUTE_CHANGE },
        AttributeChangeNotification.class.getName(),
        "Notifies a change in attribute's value.",
        descr2
    );
    
    modelInfo[size-2] = generic;
    modelInfo[size-1] = attrChange;
    
    return modelInfo; 
  }

  
  // PersistentMBean interface
  
  public void load() throws MBeanException, 
      InstanceNotFoundException {

   // throw new UnsupportedOperationException();
   if (metadata == null)
     return;
     
   Descriptor d = metadata.getMBeanDescriptor();
   String dir   = (String)d.getFieldValue(PERSIST_LOCATION);
   String file  = (String)d.getFieldValue(PERSIST_NAME);
   
   if (file != null) {
     try {
       File f = new File(dir, file);
       FileInputStream fis = new FileInputStream(f);
       ObjectInputStream ois = new ObjectInputStream(fis);
     
       metadata = (ModelMBeanInfoSupport)ois.readObject();
     }
     catch (Exception e) {
       System.out.println("Error loading MBean state");
     }
   }
  }
  
  public void store() throws MBeanException,
      InstanceNotFoundException {

   try {
     Descriptor d = metadata.getMBeanDescriptor();
     String dir  = (String)d.getFieldValue(PERSIST_LOCATION);
     String file = (String)d.getFieldValue(PERSIST_NAME);
    
     File f = new File(dir, file);
     FileOutputStream fos   = new FileOutputStream(f);
     ObjectOutputStream oos = new ObjectOutputStream(fos);
    
     oos.writeObject(metadata);
   }
   catch (IOException e) {
     throw new MBeanException(e, "Error in persisting MBean.");
   }
    //throw new UnsupportedOperationException();
  }       
            
      
  // MBeanRegistration interface
    
  public ObjectName preRegister(MBeanServer server,
                                ObjectName name) 
                                throws Exception {    

    // store the server reference
    this.server = server;
    this.name    = name;
    
    // provide a default object name
    if (name == null || name.equals("")) {
      name = new ObjectName(
          server.getDefaultDomain() +
          ":type=XMBean,name=Default"
      );
    }    

    // check if configured to load xml definition
//    if (metadata instanceof XMBeanInfo) {
//      String url = 
//          (String)metadata.getMBeanDescriptor().
//              getFieldValue(XML_DEFINITION_URL);
    
   //   Descriptor   mbeanDescriptor = metadata.getMBeanDescriptor();
   //   Descriptor[] attrDescriptors = metadata.getDescriptors("attribute");
   //   Descriptor[] operDescriptors = metadata.getDescriptors("operation");
      
//      MetaDataBuilder builder = new XMLMetaDataBuilder(
//          resource.getClass().getName(),
//          new URL(url)
//      );
//      metadata = builder.build();
      
     // metadata.setMBeanDescriptor(mbeanDescriptor);
     // metadata.setDescriptors(attrDescriptors);
//    }
  
    // create attribute and operation maps
    attributeMap = createAttributeMap(metadata.getAttributes());
    operationMap = createOperationMap(metadata.getOperations());

    
//    init();
    
    return name;
  }

  private Map createAttributeMap(MBeanAttributeInfo[] attributes) 
      throws MBeanException  {
    
    Map attrMap = new HashMap();
    
    for (int i = 0; i < attributes.length; ++i) {
      String name = attributes[i].getName();
      
      ModelMBeanAttributeInfo info = 
          (ModelMBeanAttributeInfo)attributes[i];
          
      attrMap.put(name, new XMBeanAttribute(this, info));        
    }
    
    return attrMap;
  }
  
  private Map createOperationMap(MBeanOperationInfo[] operations) 
      throws MBeanException {
      
    Map operMap = new HashMap();
    
    for (int i = 0; i < operations.length; ++i) {
      String name = operations[i].getName();
      MBeanParameterInfo[] params = operations[i].getSignature();
      
      for (int j = 0; j < params.length; ++j)
        name += params[j].getType();
        
      ModelMBeanOperationInfo info =
          (ModelMBeanOperationInfo)operations[i];
          
      XMBeanOperation operation = new XMBeanOperation(this, info);
      operMap.put(name, operation);
      
    }
    
    return operMap;
  }


  
  public void postRegister(Boolean registrationSuccessful) {
//    if (registrationSuccessful.booleanValue() == false) {
//      System.out.println("Registration of " + name + " failed.");
//      return;
//    }
//    
//    start();      
//
// For Chapter 10:
//
//    try {
//      Descriptor descr = metadata.getMBeanDescriptor();
//      Object policy = descr.getFieldValue(EXPORT);
//    
//      if (policy != null && policy instanceof ExportPolicy) {
//        ObjectName exportManager = 
//            new ObjectName("Service:name=ExportManager");        
//      
//        server.invoke(
//            exportManager,
//            "export",
//            new Object[] { name, "MyMBean"/*policy*/ },
//            new String[] { "javax.management.ObjectName", "java.lang.String" /*"book.jmx.examples.ExportPolicy"*/ }
//        );
//      }
//    }
//    catch (JMException e) {
//      e.printStackTrace();
//    }
//    
  }
  
  public void preDeregister() throws Exception {  }
  public void postDeregister() {  }
  
  
  // DynamicMBean interface
  
  public Object getAttribute(String attribute)
      throws AttributeNotFoundException, MBeanException,
      ReflectionException {
       
    XMBeanAttribute attr = 
        (XMBeanAttribute)attributeMap.get(attribute);
        
    if (attr == null)
      throw new AttributeNotFoundException();
      
    return attr.getValue();
  }

            
  public AttributeList getAttributes(String[] attributes) {
    
    if (attributes == null)
      throw new IllegalArgumentException("null array");
      
    AttributeList list = new AttributeList();
    
    for (int i = 0; i < attributes.length; ++i) {
      try {
        list.add(new Attribute(
            attributes[i], getAttribute(attributes[i])
        ));
      }
      catch (JMException ignored) {
        // if the attribute could not be retrieved, skip it            
      }
    }
    
    return list;
  }

   
  public void setAttribute(Attribute attribute) 
      throws AttributeNotFoundException,
      InvalidAttributeValueException, MBeanException,
      ReflectionException {
  
    String attrName  = attribute.getName();
    Object attrValue = attribute.getValue();

    XMBeanAttribute attr = 
        (XMBeanAttribute)attributeMap.get(attrName);
    
    if (attr == null)
      throw new AttributeNotFoundException();
    
    try {
      attr.setValue(attrValue);
    }
    catch (InstanceNotFoundException e) {
      
      // may be thrown by PersistentMBean.store()
        
      throw new MBeanException(e);
    }              
  }
            
  public AttributeList setAttributes(AttributeList list) {
      
    if (list == null)
      throw new IllegalArgumentException("null list");
      
    AttributeList results = new AttributeList();
    Iterator it           = list.iterator();
    
    while (it.hasNext()) {
      try {
        Attribute attr = (Attribute)it.next();
        setAttribute(attr);
        results.add(attr);
      }
      catch (JMException ignored) {
        // if unable to set the attribute, skip it
      }
    }    
    
    return results;
  }
          
    
  public Object invoke(String actionName, Object[] params,
                       String[] signature)
      throws MBeanException, ReflectionException {
          
    
    String method = actionName/* + "("*/;
    
    if (signature != null) {
      for (int i = 0; i < signature.length; ++i)
        method += signature[i];
    }
    
    
    XMBeanOperation operation =
        (XMBeanOperation)operationMap.get(method);
    
    if (operation == null) {
      throw new ReflectionException(
         new IllegalArgumentException("unknown operation")
      );
    }
      
    return operation.invoke(params);
  }
  
  public MBeanInfo getMBeanInfo() {
    return (MBeanInfo)metadata;
  }
    

  
}
